                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 8: lexical addressing   */
                     /*-----------------------------------*/
                     /*              evaluate             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipMain.h"

#include "SlipEnvironment.h"
#include "SlipEvaluate.h"
#include "SlipThread.h"

/*------------------------------ private macros ------------------------------*/

#define chunk_size(TYP)                                                        \
  ((sizeof(TYP) - sizeof(CNT))/ sizeof(CEL_type))

/*------------------------------ private constants ---------------------------*/

static const TXT_type BRQ_error_string = "boolean value required";
static const TXT_type ILT_error_string = "illegal expression type";
static const TXT_type ITA_error_string = "improperly terminated application";
static const TXT_type IXT_error_string = "invalid expression type";
static const TXT_type PNR_error_string = "procedure or native required";
static const TXT_type TFO_error_string = "too few operands in application";
static const TXT_type TMO_error_string = "too many operands in application";

/*------------------------------ private prototypes --------------------------*/

static EXP_type        evaluate_bindings(PRC_type,
                                         VEC_type);
static EXP_type            evaluate_body(VEC_type,
                                         VEC_type,
                                         VEC_type);
static EXP_type      evaluate_expression(EXP_type);
static EXP_type     evaluate_native_call(NAT_type,
                                         VEC_type);
static EXP_type        evaluate_sequence(VEC_type);
static EXP_type evaluate_vararg_bindings(PRZ_type,
                                         VEC_type);

/*------------------------------ private functions ---------------------------*/

static EXP_type continue_with(EXP_type Value)
  { CNT_type continuation;
    CCC_type function;
    CFN_type c_function;
    continuation = Thread_Peek();
    c_function = continuation->cfn;
    function = c_function->ccc;
    return function(Value); }

static SGN_type list_length(PAI_type List)
  { PAI_type pair;
    SGN_type count;
    count = 0;
    for (pair = List;
         is_PAI(pair);
         pair = pair->cdr)
      count++;
    if (is_NUL(pair))
      return count;
    return -1; }

/*----------------------------------------------------------------------------*/
/*------------------------------------ application ---------------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Application_String = "application";

static CFN_type Continue_application;

typedef struct aPL * aPL_type;
typedef struct aPL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     VEC_type opd; } aPL;

static const UNS_type aPL_size = chunk_size(aPL);

static EXP_type continue_application(EXP_type Procedure)
  { aPL_type application_thread;
    TAG_type tag;
    VEC_type operands;
    application_thread = (aPL_type)Thread_Peek();
    operands = application_thread->opd;
    tag = Tag_of(Procedure);
    switch (tag)
      { case NAT_tag:
              return evaluate_native_call(Procedure,
                                          operands); 
        case PRC_tag:
                 return evaluate_bindings(Procedure,
                                          operands);  
        case PRZ_tag:
          return evaluate_vararg_bindings(Procedure,
                                          operands); }
    return Main_Error_Text(PNR_error_string,
                           Application_String); }
                           
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_application(APL_type Application)
  { aPL_type application_thread;
    EXP_type expression;
    VEC_type operands;
    expression = Application->exp;
    operands   = Application->opr;
    application_thread = (aPL_type)Thread_Push(Continue_application,
                                               aPL_size);
    application_thread->opd = operands;
    return evaluate_expression(expression); }

static NIL_type initialize_application(NIL_type)
  { Continue_application = make_CFN(continue_application); }
    
/*----------------------------------------------------------------------------*/
/*------------------------------------ begin ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_begin(BEG_type Begin)
  { VEC_type sequence;
    sequence = Begin->seq;
    return evaluate_sequence(sequence); }
      
/*----------------------------------------------------------------------------*/
/*---------------------------------- bindings --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_bindings;

typedef struct bND * bND_type;
typedef struct bND { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PRC_type prc;
                     VEC_type opd;
                     VEC_type frm;
                     NBR_type pos; } bND;

static const UNS_type bND_size = chunk_size(bND);

static EXP_type continue_bindings(EXP_type Value)
  { bND_type binding_thread;
    EXP_type expression;
    NBR_type c_param_count,
             position;
    PRC_type procedure;
    VEC_type body,
             operands;
    UNS_type index,
             param_count;
    VEC_type environment,
             frame;
    binding_thread = (bND_type)Thread_Peek();
    procedure = binding_thread->prc;
    frame     = binding_thread->frm;
    position  = binding_thread->pos;
    index = position->lng;
    frame[index] = Value;
    c_param_count = procedure->par;
    param_count = c_param_count->lng;
    if (index < param_count)
      { binding_thread = (bND_type)Thread_Keep();
        position = make_NBR(++index);
        operands = binding_thread->opd;
        binding_thread->pos = position;
        expression = operands[index];
        return evaluate_expression(expression); }
    body        = procedure->bod;
    environment = procedure->env;
    return evaluate_body(body,
                         environment,
                         frame); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_bindings(PRC_type Procedure,
                                  VEC_type Operands)
  { bND_type binding_thread;
    EXP_type expression;
    NBR_type c_frame_size,
             c_param_count;
    UNS_type frame_size,
             operand_count,
             param_count;
    VEC_type body,
             environment,
             frame;
    c_param_count = Procedure->par;
    param_count = c_param_count->lng;
    operand_count = size_VEC(Operands);
    if (param_count > operand_count)
      return Main_Error_Procedure(TFO_error_string,
                                  Procedure);
    if (param_count < operand_count)
      return Main_Error_Procedure(TMO_error_string,
                                  Procedure);
    c_frame_size = Procedure->siz;
    frame_size = c_frame_size->lng;
    if (frame_size == 0)
      { body        = Procedure->bod;
        environment = Procedure->env;
        return evaluate_body(body,
                             environment,
                             Main_Empty_Vector); }
    frame = make_VEC(frame_size);
    if (param_count == 0)
      { body        = Procedure->bod;
        environment = Procedure->env;
        return evaluate_body(body,
                             environment,
                             frame); }
    binding_thread = (bND_type)Thread_Poke(Continue_bindings,
                                           bND_size);
    binding_thread->prc = Procedure;
    binding_thread->opd = Operands;
    binding_thread->frm = frame;
    binding_thread->pos = Main_One;
    expression = Operands[1];
    return evaluate_expression(expression); }

static NIL_type initialize_bindings(NIL_type)
  { Continue_bindings = make_CFN(continue_bindings); }

/*----------------------------------------------------------------------------*/
/*-------------------------------------- body --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_body;

typedef struct bOD * bOD_type;
typedef struct bOD { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     VEC_type env;
                     VEC_type frm; } bOD;

static const UNS_type bOD_size = chunk_size(bOD);

static EXP_type continue_body(EXP_type Value)
  { bOD_type body_thread;
    VEC_type environment,
             frame;
    body_thread = (bOD_type)Thread_Pop();
    environment = body_thread->env;
    frame       = body_thread->frm;
    Environment_Set_Environment_And_Frame(environment,
                                          frame); 
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_body(VEC_type Body,
                              VEC_type Environment,
                              VEC_type Frame)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Poke(Continue_body,
                                        bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(Environment,
                                          Frame); 
    return evaluate_sequence(Body); }

static EXP_type evaluate_direct_body(VEC_type Body,
                                     VEC_type Environment,
                                     VEC_type Frame)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Push(Continue_body,
                                        bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(Environment,
                                          Frame); 
    return evaluate_sequence(Body); }

static EXP_type evaluate_inline_poke(EXP_type Expression,
                                     UNS_type Frame_size)
  { bOD_type body_thread;
    VEC_type environment,
             frame;
    environment = Environment_Grow_Environment();
    frame = make_VEC(Frame_size);
    body_thread = (bOD_type)Thread_Poke(Continue_body,
                                        bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(environment,
                                          frame); 
    return evaluate_expression(Expression); }

static EXP_type evaluate_inline_push(EXP_type Expression,
                                     UNS_type Frame_size)
  { bOD_type body_thread;
    VEC_type environment,
             frame;
    environment = Environment_Grow_Environment();
    frame = make_VEC(Frame_size);
    body_thread = (bOD_type)Thread_Push(Continue_body,
                                        bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(environment,
                                          frame); 
    return evaluate_expression(Expression); }

static EXP_type evaluate_inline_sequence(VEC_type Body,
                                         UNS_type Frame_size)
  { bOD_type body_thread;
    VEC_type environment,
             frame;
    environment = Environment_Grow_Environment();
    frame = make_VEC(Frame_size);
    body_thread = (bOD_type)Thread_Push(Continue_body,
                                        bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(environment,
                                          frame); 
    return evaluate_sequence(Body); }

static NIL_type initialize_body(NIL_type)
  { Continue_body = make_CFN(continue_body); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ define --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_define;

typedef struct dEF * dEF_type;
typedef struct dEF { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     NBR_type ofs; } dEF;

static const UNS_type dEF_size = chunk_size(dEF);

static EXP_type continue_define(EXP_type Value)
  { dEF_type define_thread;
    NBR_type c_offset;
    UNS_type offset;
    define_thread = (dEF_type)Thread_Pop();
    c_offset = define_thread->ofs;
    offset = c_offset->lng;
    Environment_Local_Set(offset,
                          Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_define_function(DFF_type Define)
  { NBR_type c_frame_size,
             c_offset,
             c_param_count;
    PRC_type procedure;
    SYM_type name;
    UNS_type offset;
    VEC_type body,
             environment;
    name          = Define->nam;
    c_offset      = Define->ofs;
    c_param_count = Define->par;
    c_frame_size  = Define->siz;
    body          = Define->bod;
    environment   = Environment_Grow_Environment(); 
    procedure = make_PRC(name,
                         c_param_count,
                         c_frame_size,
                         body,
                         environment);
    offset = c_offset->lng;
    Environment_Local_Set(offset,
                          procedure);
    return procedure; }
    
static EXP_type evaluate_define_function_vararg(DFZ_type Define)
  { NBR_type c_frame_size,
             c_offset,
             c_param_count;
    PRZ_type procedure;
    SYM_type name;
    UNS_type offset;
    VEC_type body,
             environment;
    name          = Define->nam;
    c_offset      = Define->ofs;
    c_param_count = Define->par;
    c_frame_size  = Define->siz;
    body          = Define->bod;
    environment   = Environment_Grow_Environment(); 
    procedure = make_PRZ(name,
                         c_param_count,
                         c_frame_size,
                         body,
                         environment);
    offset = c_offset->lng;
    Environment_Local_Set(offset,
                          procedure);
    return procedure; }
            
static EXP_type evaluate_define_variable(DFV_type Define)
  { dEF_type define_thread;
    EXP_type expression;
    NBR_type c_offset;
    c_offset   = Define->ofs;
    expression = Define->exp;
    define_thread = (dEF_type)Thread_Push(Continue_define,
                                          dEF_size);
    define_thread->ofs = c_offset;
    return evaluate_expression(expression); }
    
static NIL_type initialize_define(NIL_type)
  { Continue_define = make_CFN(continue_define); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ direct_application --------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Direct_Application_String = "direct application";

static EXP_type evaluate_direct_application(EXP_type Value,
                                            PAI_type Arguments)
  { EXP_type value;
    FUN_type function;
    NAT_type native;
    NBR_type c_frame_size,
             c_param_count;
    PAI_type arguments;
    PRC_type procedure;
    PRZ_type procedure_vararg;
    SGN_type length;
    TAG_type tag;
    TXT_type name;
    UNS_type frame_size,
             index,
             param_count;
    VEC_type body,
             environment,
             frame,
             vector;
    tag = Tag_of(Value);
    switch (tag)
      { case NAT_tag:
          native = (NAT_type)Value;
          function = native->fun;
          length = list_length(Arguments);
          switch (length)
            { case 0:
                vector = Main_Empty_Vector;
                break;
              case -1:
                { name = native->nam;
                  return Main_Error_Text(ITA_error_string,
                                         name); }
              default:
                { arguments = Arguments;
                  vector = make_VEC(length);
                  for (index = 1;
                       index <= length;
                       index++)
                    { value = arguments->car;
                      arguments = arguments->cdr;
                      vector[index] = value; }
                  break; }}
          return function(vector); 
        case PRC_tag:
          arguments = Arguments;
          procedure = (PRC_type)Value;
          c_param_count = procedure->par;
          c_frame_size  = procedure->siz;
          environment   = procedure->env;
          param_count = c_param_count->lng;
          frame_size = c_frame_size->lng;
          frame = make_VEC(frame_size);
          for (index = 1; 
               index <= param_count;
               index++)
            { if (is_NUL(arguments))
                return Main_Error_Procedure(TFO_error_string, 
                                            procedure);
              value = arguments->car;
              arguments = arguments->cdr;
              frame[index] = value; }
          if (!is_NUL(arguments))
            return Main_Error_Procedure(TMO_error_string,
                                        procedure);
          body = procedure->bod;
          return evaluate_direct_body(body,
                                      environment,
                                      frame); 
        case PRZ_tag:
          arguments = Arguments;
          procedure_vararg = (PRZ_type)Value;
          c_param_count = procedure_vararg->par;
          c_frame_size  = procedure_vararg->siz;
          environment   = procedure_vararg->env;
          param_count = c_param_count->lng;
          frame_size = c_frame_size->lng;
          frame = make_VEC(frame_size);
          for (index = 1; 
               index <= param_count;
               index++)
            { if (is_NUL(arguments))
                return Main_Error_Procedure_Vararg(TFO_error_string, 
                                                   procedure_vararg);
              value = arguments->car;
              arguments = arguments->cdr;
              frame[index] = value;  }
          frame[param_count + 1] = arguments; 
          body = procedure_vararg->bod;
          return evaluate_direct_body(body,
                                      environment,
                                      frame); }
    return Main_Error_Text(PNR_error_string,
                           Direct_Application_String); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ global variable -----------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_global_variable(GLB_type Global)
  { EXP_type value;
    NBR_type c_offset,
             c_scope;
    UNS_type scope,
             offset;
    c_scope  = Global->scp;
    c_offset = Global->ofs;
    scope = c_scope->lng;
    offset = c_offset->lng;
    value = Environment_Global_Get(scope,
                                   offset);
    return value; }

/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_if_double;

typedef struct iFD * iFD_type;
typedef struct iFD { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type cns;
                     NBR_type csz;
                     EXP_type alt;
                     NBR_type asz; } iFD;

static const UNS_type iFD_size = chunk_size(iFD);

static EXP_type continue_if_double(EXP_type Boolean)
  { iFD_type if_thread;
    EXP_type alternative,
             consequent;
    NBR_type c_alternative_size,
             c_consequent_size;
    TAG_type tag;
    UNS_type alternative_size,
             consequent_size;
    tag = Tag_of(Boolean);
    if_thread = (iFD_type)Thread_Peek();
    switch (tag)
      { case FLS_tag:
          alternative        = if_thread->alt;
          c_alternative_size = if_thread->asz;
          alternative_size = c_alternative_size->lng;
          return evaluate_inline_poke(alternative,
                                      alternative_size);
        case TRU_tag:
          consequent        = if_thread->cns;
          c_consequent_size = if_thread->csz;
          consequent_size = c_consequent_size->lng;
          return evaluate_inline_poke(consequent,
                                      consequent_size); }
    return Main_Error_Text(BRQ_error_string,
                           Main_If_String); }

/*----------------------------------------------------------------------------*/

static CFN_type Continue_if_single;

typedef struct iFS * iFS_type;
typedef struct iFS { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type cns;
                     NBR_type csz; } iFS;

static const UNS_type iFS_size = chunk_size(iFS);

static EXP_type continue_if_single(EXP_type Boolean)
  { iFS_type if_thread;
    EXP_type consequent;
    NBR_type c_consequent_size;
    TAG_type tag;
    UNS_type consequent_size;
    tag = Tag_of(Boolean);
    if_thread = (iFS_type)Thread_Peek();
    switch (tag)
      { case FLS_tag:
          Thread_Zap();
          return Main_Unspecified;
        case TRU_tag:
          consequent        = if_thread->cns;
          c_consequent_size = if_thread->csz;
          consequent_size = c_consequent_size->lng;
          return evaluate_inline_poke(consequent,
                                      consequent_size); }
    return Main_Error_Text(BRQ_error_string,
                           Main_If_String); }
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_if_double(IFF_type If)
  { iFD_type if_thread;
    EXP_type alternative,
             consequent,
             predicate;
    NBR_type c_alternative_size,
             c_consequent_size;
    predicate          = If->prd;
    consequent         = If->cns;
    c_consequent_size  = If->csz;
    alternative        = If->alt;
    c_alternative_size = If->asz;
    if_thread = (iFD_type)Thread_Push(Continue_if_double,
                                      iFD_size);
    if_thread->cns = consequent;
    if_thread->csz = c_consequent_size;
    if_thread->alt = alternative;
    if_thread->asz = c_alternative_size;
    return evaluate_expression(predicate); }
    
static EXP_type evaluate_if_single(IFZ_type If)
  { iFS_type if_thread;
    EXP_type consequent,
             predicate;
    NBR_type c_consequent_size;
    predicate         = If->prd;
    consequent        = If->cns;
    c_consequent_size = If->csz;
    if_thread = (iFS_type)Thread_Push(Continue_if_single,
                                      iFS_size);
    if_thread->cns = consequent;
    if_thread->csz = c_consequent_size;
    return evaluate_expression(predicate); }
    
static NIL_type initialize_if(NIL_type)
  { Continue_if_double = make_CFN(continue_if_double);
    Continue_if_single = make_CFN(continue_if_single); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ lambda --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_lambda(LMB_type Lambda)
  { NBR_type c_frame_size,
             c_param_count;
    VEC_type body,
             environment;
    PRC_type procedure;
    
    c_param_count = Lambda->par;
    c_frame_size  = Lambda->siz;
    body          = Lambda->bod; 
    environment = Environment_Grow_Environment();
    procedure = make_PRC(Main_Anonymous,
                         c_param_count,
                         c_frame_size,
                         body,
                         environment);
    return procedure; }
    
static EXP_type evaluate_lambda_vararg(LMZ_type Lambda)
  { NBR_type c_frame_size,
             c_param_count;
    VEC_type body,
             environment;
    PRZ_type procedure;
    
    c_param_count = Lambda->par;
    c_frame_size  = Lambda->siz;
    body          = Lambda->bod; 
    environment = Environment_Grow_Environment();
    procedure = make_PRZ(Main_Anonymous,
                         c_param_count,
                         c_frame_size,
                         body,
                         environment);
    return procedure; }
    
/*----------------------------------------------------------------------------*/
/*------------------------------------ local variable ------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_local_variable(LCL_type Local)
  { EXP_type value;
    NBR_type c_offset;
    UNS_type offset;
    c_offset = Local->ofs;
    offset = c_offset->lng;
    value = Environment_Local_Get(offset);
    return value; }

/*----------------------------------------------------------------------------*/
/*-------------------------------- native call -------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_native_call;

typedef struct nCL * nCL_type;
typedef struct nCL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     VEC_type opd;
                     VEC_type vec;
                     NAT_type nat;
                     NBR_type pos; } nCL;

static const UNS_type nCL_size = chunk_size(nCL);

static EXP_type continue_native_call(EXP_type Value)
  { nCL_type native_call_thread;
    EXP_type expression;
    FUN_type function;
    NAT_type native;
    NBR_type position;
    UNS_type index,
             size_v;
    VEC_type operands,
             vector;
    native_call_thread = (nCL_type)Thread_Peek();
    vector   = native_call_thread->vec;
    position = native_call_thread->pos;
    index = position->lng;
    size_v = size_VEC(vector);
    vector[index] = Value;
    if (index++ < size_v)
      { native_call_thread = (nCL_type)Thread_Keep();
        operands = native_call_thread->opd;
        expression = operands[index];
        position = make_NBR(index);
        native_call_thread->pos = position;
        return evaluate_expression(expression); }
    native = native_call_thread->nat;
    function = native->fun;
    Thread_Zap();
    return function(vector); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_native_call(NAT_type Native,
                                     VEC_type Operands)
  { nCL_type native_call_thread;
    EXP_type expression;
    FUN_type function;
    UNS_type size_o;
    VEC_type vector;
    size_o = size_VEC(Operands);
    if (size_o == 0)
      { function = Native->fun;
        Thread_Zap();
        return function(Main_Empty_Vector); }
    vector = make_VEC(size_o);
    native_call_thread = (nCL_type)Thread_Poke(Continue_native_call,
                                               nCL_size);
    native_call_thread->opd = Operands;
    native_call_thread->vec = vector;
    native_call_thread->nat = Native;
    native_call_thread->pos = Main_One;
    expression = Operands[1];
    return evaluate_expression(expression); }
                       
static NIL_type initialize_native_call(NIL_type)
  { Continue_native_call = make_CFN(continue_native_call); }
  
/*----------------------------------------------------------------------------*/
/*------------------------------------ quote ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_quote(QUO_type Quote)
  { EXP_type expression;
    expression = Quote->exp;
    return expression; }        /* clone!!! */

/*----------------------------------------------------------------------------*/
/*---------------------------------- sequence --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_sequence;

typedef struct sEQ * sEQ_type;
typedef struct sEQ { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     VEC_type exs;
                     NBR_type pos; } sEQ;

static const UNS_type sEQ_size = chunk_size(sEQ);

static EXP_type continue_sequence(EXP_type Value)
  { sEQ_type sequence_thread;
    EXP_type expression;
    VEC_type expressions;
    NBR_type position;
    UNS_type index,
             size_x;
    sequence_thread = (sEQ_type)Thread_Peek();
    expressions = sequence_thread->exs;
    position    = sequence_thread->pos;
    index = position->lng + 1;
    expression = expressions[index];
    size_x = size_VEC(expressions);
    if (index < size_x)
      { sequence_thread = (sEQ_type)Thread_Keep();
        position = make_NBR(index);
        sequence_thread->pos = position; }
    else
      Thread_Zap();
    return evaluate_expression(expression); }
    
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_sequence(VEC_type Expressions)
  { sEQ_type sequence_thread;
    EXP_type expression;
    UNS_type size_x;
    size_x = size_VEC(Expressions);
    expression = Expressions[1];
    if (size_x > 1)
      { sequence_thread = (sEQ_type)Thread_Push(Continue_sequence,
                                                sEQ_size);
        sequence_thread->exs = Expressions;
        sequence_thread->pos = Main_One; }
    return evaluate_expression(expression); }
    
static NIL_type initialize_sequence(NIL_type)
  { Continue_sequence = make_CFN(continue_sequence); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ set global ----------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_set_global;

typedef struct sTG * sTG_type;
typedef struct sTG { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     NBR_type scp;
                     NBR_type ofs; } sTG;

const static UNS_type sTG_size = chunk_size(sTG);

static EXP_type continue_set_global(EXP_type Value)
  { sTG_type set_thread;
    NBR_type c_offset,
             c_scope;
    UNS_type offset,
             scope;
    set_thread = (sTG_type)Thread_Pop();
    c_scope  = set_thread->scp;
    c_offset = set_thread->ofs;
    scope = c_scope->lng;
    offset = c_offset->lng;
    Environment_Global_Set(scope,
                           offset,
                           Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_set_global(STG_type Set)
  { sTG_type set_thread;
    EXP_type expression;
    NBR_type c_offset,
             c_scope;
    c_scope    = Set->scp;
    c_offset   = Set->ofs;
    expression = Set->exp;
    set_thread = (sTG_type)Thread_Push(Continue_set_global,
                                       sTG_size);
    set_thread->scp = c_scope;
    set_thread->ofs = c_offset;
    return evaluate_expression(expression); }
    
static NIL_type initialize_evaluate_set_global(NIL_type)
  { Continue_set_global = make_CFN(continue_set_global); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ set local -----------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_set_local;

typedef struct sTL * sTL_type;
typedef struct sTL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     NBR_type ofs; } sTL;

const static UNS_type sTL_size = chunk_size(sTL);

static EXP_type continue_set_local(EXP_type Value)
  { sTL_type set_thread;
    NBR_type c_offset;
    UNS_type offset;
    set_thread = (sTL_type)Thread_Pop();
    c_offset = set_thread->ofs;
    offset   = c_offset->lng;
    Environment_Local_Set(offset,
                          Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_set_local(STL_type Set)
  { sTL_type set_thread;
    EXP_type expression;
    NBR_type c_offset;
    c_offset   = Set->ofs;
    expression = Set->exp;
    set_thread = (sTL_type)Thread_Push(Continue_set_local,
                                       sTL_size);
    set_thread->ofs = c_offset;
    return evaluate_expression(expression); }
        
static NIL_type initialize_evaluate_set_local(NIL_type)
  { Continue_set_local = make_CFN(continue_set_local); }
  
/*----------------------------------------------------------------------------*/
/*------------------------------------ value ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_value(EXP_type Value)
  { return Value; }

/*----------------------------------------------------------------------------*/
/*------------------------------- vararg_bindings ----------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_vararg,
                Continue_vararg_bindings;

typedef struct vRG * vRG_type;
typedef struct vRG { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PRZ_type prc;
                     VEC_type frm;
                     VEC_type opd;
                     PAI_type lst;
                     NBR_type pos; } vRG;

static const UNS_type vRG_size = chunk_size(vRG);

static EXP_type continue_vararg(EXP_type Value)
  { vRG_type binding_thread;
    EXP_type expression;
    NBR_type c_param_count,
             position;
    PAI_type list;
    PRZ_type procedure;
    VEC_type environment,
             frame,
             operands;
    UNS_type index,
             operand_size,
             param_count;
    VEC_type body;
    binding_thread = (vRG_type)Thread_Peek();
    operands = binding_thread->opd;
    list     = binding_thread->lst;
    position = binding_thread->pos;
    index = position->lng;
    operand_size = size_VEC(operands);
    list = make_PAI(Value,
                    list);                             
    if (index < operand_size)
      { binding_thread = (vRG_type)Thread_Keep();
        position = make_NBR(++index);
        binding_thread->lst = list;
        binding_thread->pos = position;
        expression = operands[index];
        return evaluate_expression(expression); }
    procedure = binding_thread->prc;
    frame     = binding_thread->frm;
    c_param_count = procedure->par;
    body          = procedure->bod;
    environment   = procedure->env;
    param_count = c_param_count->lng;
    list = Main_Reverse(list);
    frame[param_count + 1] = list;
    return evaluate_body(body,
                         environment,
                         frame); }

static EXP_type continue_vararg_bindings(EXP_type Value)
  { vRG_type binding_thread;
    EXP_type expression;
    NBR_type c_param_count,
             position;
    PRZ_type procedure;
    VEC_type environment,
             frame,
             operands;
    UNS_type index,
             operand_count,
             param_count;
    VEC_type body;
    binding_thread = (vRG_type)Thread_Peek();
    procedure = binding_thread->prc;
    frame     = binding_thread->frm;
    operands  = binding_thread->opd;
    position  = binding_thread->pos;
    index = position->lng;
    frame[index++] = Value;
    operand_count = size_VEC(operands);
    if (index > operand_count)
      { body        = procedure->bod;
        environment = procedure->env;
        frame[index] = Main_Empty_Pair;
        return evaluate_body(body,
                             environment,
                             frame); }
    c_param_count = procedure->par;
    param_count = c_param_count->lng;
    if (index > param_count)
      binding_thread = (vRG_type)Thread_Patch(Continue_vararg);
    position = make_NBR(index);
    binding_thread->lst = Main_Empty_Pair;
    binding_thread->pos = position;
    expression = operands[index];
    return evaluate_expression(expression); }
    
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_vararg_bindings(PRZ_type Procedure,
                                         VEC_type Operands)
  { vRG_type binding_thread;
    CFN_type c_function;
    EXP_type expression;
    NBR_type c_frame_size,
             c_param_count;
    UNS_type frame_size,
             operand_count,
             param_count;
    VEC_type body,
             environment,
             frame;
    c_param_count = Procedure->par;
    param_count = c_param_count->lng;
    operand_count = size_VEC(Operands);
    if (param_count > operand_count)
      return Main_Error_Procedure_Vararg(TFO_error_string,
                                         Procedure);
    c_frame_size = Procedure->siz;
    frame_size = c_frame_size->lng;
    frame = make_VEC(frame_size);
    if (operand_count == 0)
      { body        = Procedure->bod;
        environment = Procedure->env;
        frame[1] = Main_Empty_Pair;
        return evaluate_body(body,
                             environment,
                             frame); }
    if (param_count == 0)
      c_function = Continue_vararg;
    else
      c_function = Continue_vararg_bindings;
    binding_thread = (vRG_type)Thread_Poke(c_function,
                                           vRG_size);
    binding_thread->prc = Procedure;
    binding_thread->frm = frame;
    binding_thread->opd = Operands;
    binding_thread->lst = Main_Empty_Pair;
    binding_thread->pos = Main_One;
    expression = Operands[1];
    return evaluate_expression(expression); }
                
static NIL_type initialize_vararg_bindings(NIL_type)
  { Continue_vararg = make_CFN(continue_vararg);
    Continue_vararg_bindings = make_CFN(continue_vararg_bindings); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ while ---------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_while_body,
                Continue_while_predicate;

typedef struct wHI * wHI_type;
typedef struct wHI { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type prd;
                     NBR_type psz;
                     VEC_type bod;
                     NBR_type bsz;
                     EXP_type res; } wHI;

static const UNS_type wHI_size = chunk_size(wHI);

static EXP_type continue_while_predicate(EXP_type Boolean)
  { wHI_type while_thread;
    EXP_type result;
    NBR_type c_body_size;
    UNS_type body_size;
    VEC_type body;
    while_thread = (wHI_type)Thread_Peek();
    if (is_FLS(Boolean))
      { result = while_thread->res;
        Thread_Zap();
        return result; }
    while_thread = (wHI_type)Thread_Patch(Continue_while_body);
    body = while_thread->bod;
    c_body_size = while_thread->bsz;
    body_size = c_body_size->lng;
    return evaluate_inline_sequence(body,
                                    body_size); }

static EXP_type continue_while_body(EXP_type Value)
  { wHI_type while_thread;
    EXP_type predicate;
    NBR_type c_predicate_size;
    UNS_type predicate_size;
    while_thread = (wHI_type)Thread_Peek();
    predicate        = while_thread->prd;
    c_predicate_size = while_thread->psz;
    while_thread = (wHI_type)Thread_Patch(Continue_while_predicate);
    while_thread->res = Value;
    predicate_size = c_predicate_size->lng;
    return evaluate_inline_push(predicate,
                                predicate_size); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_while(WHI_type While)
  { wHI_type while_thread;
    EXP_type predicate;
    NBR_type c_body_size,
             c_predicate_size;
    UNS_type predicate_size;
    VEC_type body;
    predicate        = While->prd;
    c_predicate_size = While->psz;
    body             = While->bod;
    c_body_size      = While->bsz;
    while_thread = (wHI_type)Thread_Push(Continue_while_predicate,
                                         wHI_size);
    while_thread->prd = predicate;
    while_thread->psz = c_predicate_size;
    while_thread->bod = body;
    while_thread->bsz = c_body_size;
    while_thread->res = Main_Unspecified;
    predicate_size = c_predicate_size->lng;
    return evaluate_inline_push(predicate,
                                predicate_size); }
                                             
static NIL_type initialize_while(NIL_type)
  { Continue_while_predicate = make_CFN(continue_while_predicate);
    Continue_while_body = make_CFN(continue_while_body); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ expression ----------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_expression(EXP_type Expression)
  { TAG_type tag;
    tag = Tag_of(Expression);
    switch (tag)
      { case APL_tag:
                     return evaluate_application(Expression);
        case BEG_tag:
                           return evaluate_begin(Expression);
        case DFF_tag:
                 return evaluate_define_function(Expression);
        case DFV_tag:
                 return evaluate_define_variable(Expression);
        case DFZ_tag:
          return evaluate_define_function_vararg(Expression);
        case GLB_tag:
                 return evaluate_global_variable(Expression);
        case IFF_tag:
                       return evaluate_if_double(Expression);
        case IFZ_tag:
                       return evaluate_if_single(Expression);
        case LCL_tag:
                  return evaluate_local_variable(Expression);
        case LMB_tag:
                          return evaluate_lambda(Expression);
        case LMZ_tag:
                   return evaluate_lambda_vararg(Expression);
        case QUO_tag:
                           return evaluate_quote(Expression);
        case STG_tag:
                      return evaluate_set_global(Expression);
        case STL_tag:
                       return evaluate_set_local(Expression);
        case WHI_tag:
                           return evaluate_while(Expression);
        case CHA_tag:
        case FLS_tag:
        case NAT_tag:
        case NBR_tag:
        case NUL_tag:
        case PAI_tag:
        case PRC_tag:
        case PRZ_tag:
        case REA_tag:
        case STR_tag:
        case SYM_tag:
        case TRU_tag:
        case USP_tag:
        case VEC_tag:
                           return evaluate_value(Expression);
        case CFN_tag:
        case CNT_tag:
        case ENV_tag:
        case FRM_tag:
          return Main_Error_Tag(ILT_error_string, 
                                tag); }
    return Main_Error_Tag(IXT_error_string,
                          tag); }

/*------------------------------ public functions ----------------------------*/

EXP_type Evaluate_Apply(EXP_type Procedure,
                        PAI_type Arguments)
  { return evaluate_direct_application(Procedure,
                                       Arguments); }

EXP_type Evaluate_Continue(EXP_type Value)
  { return continue_with(Value); }

EXP_type Evaluate_Evaluate(EXP_type Expression)
  { return evaluate_expression(Expression); }

NIL_type Evaluate_Initialize(NIL_type)
  { initialize_application();
    initialize_bindings();
    initialize_body();
    initialize_define();
    initialize_if();
    initialize_native_call();
    initialize_sequence();
    initialize_evaluate_set_global();
    initialize_evaluate_set_local();
    initialize_vararg_bindings();
    initialize_while(); }
